// 

/**************************************************************************\
 * Copyright (c) Kongsberg Oil & Gas Technologies AS
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 * Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
\**************************************************************************/

// Pulls in __COIN_SOXT__ define used below.
#include <Inventor/Xt/SoXtBasic.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H

#include <soxtdefs.h>
#include <Inventor/Xt/SoXtObject.h>
#include <Inventor/Xt/devices/SoXtDevice.h>
#include <Inventor/Xt/SoXtComponent.h>

/*!
  \class SoXtObject Inventor/Xt/SoXtObject.h
  \brief The SoXtObject class is the common superclass for all SoXt component classes.

  \ingroup misc

  The purpose of making this class the superclass of all SoXt
  device, component and viewer classes is to be able to do runtime
  type checking of the SoXt objects.


  You can place the macro SOXT_OBJECT_HEADER(classname,parentname)
  within a class definition header for SoXt extension components to
  automatically provide the necessary definitions for setting up a
  runtime type system for your extension classes:

  \code
  #ifndef MYSPECIALVIEWER_H
  #define MYSPECIALVIEWER_H

  class MySpecialViewer : public SoXtExaminerViewer {
    SOXT_OBJECT_HEADER(MySpecialViewer, SoXtExaminerViewer);
    // [rest of class definition follows]
  };

  #endif // !MYSPECIALVIEWER_H
  \endcode

  Then put the SOXT_OBJECT_SOURCE(classname) macro within the
  actual implementation source code file to include the necessary
  code for the runtime type system:

  \code
  #include <MySpecialViewer.h>

  SOQT_OBJECT_SOURCE(MySpecialViewer);

  // [rest of class implementation]
  \endcode


  See also the documentation of the SoType class in Coin or Inventor.
 */

// *************************************************************************

/*!
  \fn SoType SoXtObject::getTypeId

  Returns the type identification of an object derived from a class
  inheriting SoXtObject.  This is used for runtime type checking
  and "downward" casting.

  Usage example:

  \code
  void foo(SoXtViewer * comp)
  {
    if (comp->getTypeId() == SoXtExaminerViewer::getClassTypeId()) {
      // safe downward cast, knows the type
      SoXtExaminerViewer * exviewer = (SoXtExaminerViewer *)comp;
      /// [then something] ///
    }
    else if (comp->getTypeId().isOfType(SoXtFlyViewer::getClassTypeId())) {
      // safe downward cast, knows the type
      SoXtFlyViewer * flyviewer = (SoXtFlyViewer *)comp;
      // then something else
    }
  }
  \endcode
*/
// FIXME: add doc above to explain how external developers can use the
// type system for their own extension classes. 20020502 mortene.

// This is a private variable.
SoType SoXtObject::classTypeId SOXT_STATIC_SOTYPE_INIT;

/*!
  Sets up initialization for data common to all instances of this
  class, submitting necessary information to the internal SoXt type
  system.
*/
void
SoXtObject::initClass(void)
{
  assert(SoXtObject::classTypeId == SoType::badType());
  SoXtObject::classTypeId =
    SoType::createType(SoType::badType(), "SoXtObject");
}

/*!
  Returns \c TRUE if the type of this object is either of the same
  type or inherited from \a type.
*/
SbBool
SoXtObject::isOfType(SoType type) const
{
  return this->getTypeId().isDerivedFrom(type);
}

/*!
  This static method returns the SoType object associated with
  objects of this class.
*/
SoType
SoXtObject::getClassTypeId(void)
{
  return SoXtObject::classTypeId;
}

/*!
  Initialize the type system of SoXtObject, all SoXt device
  classes and all SoXt components (including viewers).
*/
void
SoXtObject::init(void)
{
  SoXtObject::initClass();
  SoXtDevice::initClasses();
  SoXtComponent::initClasses();
}
